/* 
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005/2006, Anthony Minessale II <anthmct@yahoo.com>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthmct@yahoo.com>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 *
 *
 * mod_squirrelfish.c -- Squirrelfish (WebKit JavaScript Engine) Language Module
 *
 */
#include <switch.h>
#include <JavaScriptCore/JavaScript.h>

#include "mod_squirrelfish.h"
#include "mod_squirrelfish_private.h"


SWITCH_MODULE_LOAD_FUNCTION(mod_squirrelfish_load);
SWITCH_MODULE_SHUTDOWN_FUNCTION(mod_squirrelfish_shutdown);
SWITCH_MODULE_DEFINITION(mod_squirrelfish, mod_squirrelfish_load, mod_squirrelfish_shutdown, NULL);

/*
 * Globals
 */
static struct {
	switch_memory_pool_t *pool;

	/*
	 * globally shared class definitions
	 */
	JSClassRef sessionClass;
} globals;

/*
 *
 */
static JSValueRef log_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
		 size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception);

/*
 * Functions
 */
static struct globalFunction {
	const char	*name;
	JSFunctionCB_t	 func;
} globalFunctions[] = {
	{ "consoleLog",		&log_cb },
	{ "console_log",	&log_cb },
	{ "setGlobalVariable",	NULL },
	{ "getGlobalVariable",	NULL },
	{ "exit",		NULL },
	{ "include",		NULL },
	{ "bridge",		NULL },
	{ "apiExecute",		NULL },
	{ "use",		NULL },
	{ "fileDelete",		NULL },
	{ "system",		NULL },

#ifdef HAVE_CURL
	{ "fetchURL",		NULL },
	{ "fetchUrl",		NULL },
	{ "fetchURLHash",	NULL },
	{ "fetchUrlHash",	NULL },
	{ "fetchURLFile",	NULL },
	{ "fetchUrlFile",	NULL },
#endif

	{ 0, 0 }	/*!< don't delete! */
};

/**
 * init_functions
 * \brief	create global functions in context
 * \param[in]	context	the context
 * \return	SWITCH_STATUS_SUCCESS on success, something else on error
 */
static switch_status_t init_functions(JSContextRef context)
{
	struct globalFunction *gfunc = globalFunctions;
	JSObjectRef globalObject = JSContextGetGlobalObject(context);
	JSStringRef name;
	JSObjectRef function;

	if(!context || !globalObject)
		return SWITCH_STATUS_GENERR;

	while(gfunc && gfunc->name) {
		if(gfunc->func) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Creating global function %s\n", gfunc->name);

			name = JSStringCreateWithUTF8CString(gfunc->name);
			function = JSObjectMakeFunctionWithCallback(context, name, gfunc->func);
			JSObjectSetProperty(context, globalObject, name, function, kJSPropertyAttributeNone, NULL);
			JSStringRelease(name);
		}
		gfunc++;
	}

	return SWITCH_STATUS_SUCCESS;
}


/******************************************************************************
 * Session class
 ******************************************************************************/

static JSStaticValue sessionClassStaticValues[] = {
	{ 0, 0, 0, 0 }
};

static JSStaticFunction sessionClassStaticFunctions[] = {
	{ "log",	&log_cb,		kJSPropertyAttributeNone },	/*!< core logger function */
	{ 0, 0, 0 }
};

static JSClassDefinition sessionClass_definition = {
	.version	= 0,
	.attributes	= kJSClassAttributeNone,

	.className	= "Session",
	.parentClass	= NULL,

	.staticValues	= sessionClassStaticValues,	/*!< static values */
	.staticFunctions= sessionClassStaticFunctions,	/*!< static members */

	.initialize		= NULL,		/*!< invoked during creation of the object */
	.finalize		= NULL,		/*!< called before garbage collecting this object */
	.hasProperty		= sessionClass_HasProperty,	/*!< check if object has a specific property */
	.getProperty		= sessionClass_GetProperty,	/*!< get value of property */
	.setProperty		= sessionClass_SetProperty,	/*!< set value of property */
	.deleteProperty		= sessionClass_DeleteProperty,	/*!< delete property */
	.getPropertyNames	= NULL,		/*!< return names of all properties */
	.callAsFunction		= NULL, 	/*!< invoked if object has been called as function */
	.callAsConstructor	= NULL,		/*!< invoked if new has been called on class */
	.hasInstance		= NULL,		/*!< invoked on instanceof call */
	.convertToType		= NULL		/*!< convert object to string or integer */
};


/*
 * Session class callback functions
 */


/**
 * session_streamfile_cb
 * \brief	stream a file
 * \todo	DTMF input handling, proper exception handling
 */
static JSValueRef session_streamfile_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
				 size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
	JSSession *jss = JSObjectGetPrivate(thisObject);
	switch_core_session_t *session;
	switch_channel_t *channel;
	switch_file_handle_t fhandle = { 0 };
	switch_input_args_t  args = { 0 };
	char filename[PATH_MAX];
	const char *value;
	JSStringRef string;

	if(!jss) {
		return JSValueMakeBoolean(context, false);
	}

	session = jss->session;
	if(!session) {
		return JSValueMakeBoolean(context, false);
	}

	channel = switch_core_session_get_channel(session);
	if(!channel) {
		return JSValueMakeBoolean(context, false);
	}

	if(argumentCount > 0) {
		string = JSValueToStringCopy(context, arguments[0], NULL);
		JSStringGetUTF8CString(string, filename, PATH_MAX - 1);
		JSStringRelease(string);

		if(switch_file_exists(filename, NULL) != SWITCH_STATUS_SUCCESS) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "file \"%s\" not found", filename);
			return JSValueMakeBoolean(context, false);
		}
	}

	if (argumentCount > 3) {
		/* TODO: proper exception handling */
		fhandle.samples = JSValueToNumber(context, arguments[3], NULL);
	}

	if((value = switch_channel_get_variable(channel, "stream_prebuffer"))) {
		int prebuffer = atoi(value);

		if (prebuffer > 0) {
			fhandle.prebuf = prebuffer;
		}
	}

	switch_ivr_play_file(session, &fhandle, filename, &args);

	return JSValueMakeBoolean(context, true);
}


/**
 * session_set_variable_cb
 * \brief	set a channel variable method callback
 */
static JSValueRef session_set_variable_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
				 size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
	JSSession *jss = JSObjectGetPrivate(thisObject);
	switch_core_session_t *session;
	switch_channel_t *channel;

	if(!jss) {
		return JSValueMakeBoolean(context, false);
	}

	session = jss->session;
	if(!session) {
		return JSValueMakeBoolean(context, false);
	}

	channel = switch_core_session_get_channel(session);
	if(!channel) {
		return JSValueMakeBoolean(context, false);
	}

	if(argumentCount > 1) {
		char name[1024], value[1024];
		JSStringRef string;

		string = JSValueToStringCopy(context, arguments[0], NULL);
		JSStringGetUTF8CString(string, name, sizeof(name) - 1);
		JSStringRelease(string);

		string = JSValueToStringCopy(context, arguments[1], NULL);
		JSStringGetUTF8CString(string, value, sizeof(value) - 1);
		JSStringRelease(string);

		switch_channel_set_variable(channel, name, value);

		return JSValueMakeBoolean(context, true);
	}

	return JSValueMakeBoolean(context, false);
}

/**
 * session_get_variable_cb
 * \brief	get a channel variable method callback
 */
static JSValueRef session_get_variable_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
				 size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
	JSSession *jss = JSObjectGetPrivate(thisObject);
	switch_core_session_t *session;
	switch_channel_t *channel;

	if(!jss || !jss->session) {
		return JSValueMakeBoolean(context, false);
	}

	session = jss->session;
	channel = switch_core_session_get_channel(session);
	if(!channel) {
		return JSValueMakeBoolean(context, false);
	}

	if(argumentCount > 0) {
		char tmp[1024];
		const char *value;
		JSStringRef string;

		string = JSValueToStringCopy(context, arguments[0], NULL);

		JSStringGetUTF8CString(string, tmp, sizeof(tmp) - 1);
		JSStringRelease(string);

		value = switch_channel_get_variable(channel, tmp);

		if(value) {
			JSValueRef retval;

			string = JSStringCreateWithUTF8CString(value);

			retval = JSValueMakeString(context, string);
			JSStringRelease(string);

			return retval;
		}
	}

	return JSValueMakeBoolean(context, false);
}

/**
 * session_answer_cb
 * \brief	answer session method callback
 */
static JSValueRef session_answer_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
				 size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
	JSSession *jss = JSObjectGetPrivate(thisObject);
	switch_core_session_t *session;
	switch_channel_t *channel;

	if(!jss || !jss->session) {
		return JSValueMakeBoolean(context, false);
	}

	session = jss->session;
	channel = switch_core_session_get_channel(session);
	if(!channel) {
		return JSValueMakeBoolean(context, false);
	}

	switch_channel_answer(channel);

	return JSValueMakeBoolean(context, true);
}


/**
 * session_pre_answer_cb
 * \brief	pre-answer session method callback
 */
static JSValueRef session_pre_answer_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
				 size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
	JSSession *jss = JSObjectGetPrivate(thisObject);
	switch_core_session_t *session;
	switch_channel_t *channel;

	if(!jss || !jss->session) {
		return JSValueMakeBoolean(context, false);
	}

	session = jss->session;
	channel = switch_core_session_get_channel(session);
	if(!channel) {
		return JSValueMakeBoolean(context, false);
	}

	switch_channel_pre_answer(channel);

	return JSValueMakeBoolean(context, true);
}

/**
 * session_hangup_cb
 * \brief	hangup session method callback
 */
static JSValueRef session_hangup_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
				 size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
	JSSession *jss = JSObjectGetPrivate(thisObject);
	switch_core_session_t *session;
	switch_channel_t *channel;
	switch_call_cause_t cause = SWITCH_CAUSE_NORMAL_CLEARING;

	if(!jss || !jss->session) {
		return JSValueMakeBoolean(context, false);
	}
	session = jss->session;

	channel = switch_core_session_get_channel(session);
	if(!channel) {
		return JSValueMakeBoolean(context, false);
	}

	if(argumentCount > 0) {
		char tmp[1024];
		JSStringRef string;

		string = JSValueToStringCopy(context, arguments[0], NULL);

		JSStringGetUTF8CString(string, tmp, sizeof(tmp) - 1);
		JSStringRelease(string);

		cause = switch_channel_str2cause(tmp);
	}

	switch_channel_hangup(channel, cause);
	switch_core_session_kill_channel(session, SWITCH_SIG_KILL);

	return JSValueMakeBoolean(context, true);
}


/**
 * session_ready_cb
 * \brief	session object "ready" property getter
 */
static JSValueRef session_ready_cb(JSContextRef context, JSObjectRef thisObject)
{
	JSSession *jss = JSObjectGetPrivate(thisObject);

	if(!jss || !jss->session) {
		return JSValueMakeBoolean(context, false);
	}

	if(!switch_channel_ready(switch_core_session_get_channel(jss->session))) {
		return JSValueMakeBoolean(context, false);
	}

	return JSValueMakeBoolean(context, true);
}

/**
 * session_state_cb
 * \brief	session object "state" property getter
 */
static JSValueRef session_state_cb(JSContextRef context, JSObjectRef thisObject)
{
	JSSession *jss = JSObjectGetPrivate(thisObject);
	switch_core_session_t *session;
	switch_channel_t *channel;
	JSStringRef string;
	JSValueRef value;

	if(!jss || !jss->session) {
		return JSValueMakeUndefined(context);
	}

	session = jss->session;
	channel = switch_core_session_get_channel(session);
	if(!channel) {
		return JSValueMakeUndefined(context);
	}

	switch(switch_channel_get_state(channel)) {
	case CS_NEW:
		string = JSStringCreateWithUTF8CString("new");
		break;

	case CS_INIT:
		string = JSStringCreateWithUTF8CString("init");
		break;

	case CS_ROUTING:
		string = JSStringCreateWithUTF8CString("routing");
		break;

	case CS_SOFT_EXECUTE:
		string = JSStringCreateWithUTF8CString("soft_execute");
		break;

	case CS_EXECUTE:
		string = JSStringCreateWithUTF8CString("execute");
		break;

	case CS_EXCHANGE_MEDIA:
		string = JSStringCreateWithUTF8CString("exchange_media");
		break;

	case CS_PARK:
		string = JSStringCreateWithUTF8CString("park");
		break;

	case CS_CONSUME_MEDIA:
		string = JSStringCreateWithUTF8CString("consume_media");
		break;

	case CS_HIBERNATE:
		string = JSStringCreateWithUTF8CString("hibernate");
		break;

	case CS_RESET:
		string = JSStringCreateWithUTF8CString("reset");
		break;

	case CS_HANGUP:
		string = JSStringCreateWithUTF8CString("hangup");
		break;

	case CS_DONE:
		string = JSStringCreateWithUTF8CString("done");
		break;

	default:
		string = JSStringCreateWithUTF8CString("unknown");
		break;
	}

	value = JSValueMakeString(context, string);
	JSStringRelease(string);

	return value;
}

/******************************************************************************
 *
 ******************************************************************************/

/*
 * Array of global classes
 */
static struct globalClass {
	const char	*name;
	JSClassRef	*classRef;
} globalClasses[] = {
	{ "Session", &globals.sessionClass },
	{ 0, 0 }
};

/**
 * init_classes
 * \brief	Create global class objects (e.g. "Session")
 */
static switch_status_t init_classes(JSContextRef context)
{
	struct globalClass *gclass = globalClasses;
	JSObjectRef globalObject = JSContextGetGlobalObject(context);
	JSStringRef name;
	JSObjectRef object;

	if(!context || !globalObject)
		return SWITCH_STATUS_GENERR;

	while(gclass && gclass->name) {
		if(gclass->classRef) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Creating global class %s\n", gclass->name);

			name = JSStringCreateWithUTF8CString(gclass->name);
			object = JSObjectMake(context, *gclass->classRef, NULL);
			JSObjectSetProperty(context, globalObject, name, object, kJSPropertyAttributeNone, NULL);
			JSStringRelease(name);
		}
		gclass++;
	}

	return SWITCH_STATUS_SUCCESS;
}

/*
 * Array of session object's methods
 */
static struct objectMethod {
	const char	*name;
	JSFunctionCB_t	 func;
} sessionMethods[] = {
	{ "answer",	&session_answer_cb },
	{ "preAnswer",	&session_pre_answer_cb },
	{ "hangup",	&session_hangup_cb },
	{ "getVariable", &session_get_variable_cb },
	{ "setVariable", &session_set_variable_cb },
	{ "streamFile",	&session_streamfile_cb },
	{ 0, 0 }
};


/*
 * Array of session object's dynamic properties
 *
 * Dynamic properties are handled via session_getProperty()
 * this array is used there
 */
static struct objectProperty {
	const char		*name;
	JSPropertySetterCB_t	 setfunc;
	JSPropertyGetterCB_t	 getfunc;
	unsigned int	 	 flags;
} sessionProperties[] = {
	{ "ready", 0, &session_ready_cb, kJSPropertyAttributeReadOnly },
	{ "state", 0, &session_state_cb, kJSPropertyAttributeReadOnly },
	{ 0, 0, 0, 0 }
};


/**
 * sessionClass_HasProperty
 * \brief	Check if session object has specified property
 * \param[in]	ctx	context reference
 * \param[in]	object	session object reference
 * \param[in]	propertyName	name of the property to check
 * \param[out]	exception	storage for exception data
 * \return	true if available, false if not or error
 *
 * \todo	Read-Only static properties?
 */
static bool sessionClass_HasProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName)
{
	struct objectProperty *session_property = sessionProperties;

	if(!JSObjectGetPrivate(object))
		return false;

	while(session_property && session_property->name) {
		if(JSStringIsEqualToUTF8CString(propertyName, session_property->name)) {

			/* found property */
			return true;
		}
		session_property++;
	}

	/* TODO: need to handle read-only static values here too? (e.g. "uuid") */

	return false;
}

/**
 * sessionClass_GetProperty
 * \brief	Get property handler of session object
 * \param[in]	ctx	context reference
 * \param[in]	object	session object reference
 * \param[in]	propertyName	name of the property to get
 * \param[out]	exception	storage for exception data
 * \return	reference to value on success, NULL on failure
 *
 * \todo	Read-Only static properties?
 */
static JSValueRef sessionClass_GetProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef* exception)
{
	struct objectProperty *session_property = sessionProperties;

	if(!JSObjectGetPrivate(object))
		return NULL;

	while(session_property && session_property->name) {
		if(JSStringIsEqualToUTF8CString(propertyName, session_property->name)) {

			/* found property */
			return session_property->getfunc(ctx, object);
		}
		session_property++;
	}

	/* TODO: need to handle read-only static values here too? (e.g. "uuid") */

	return NULL;
}


/**
 * sessionClass_SetProperty
 * \brief	Set property handler of session object
 * \param[in]	ctx	context reference
 * \param[in]	object	session object reference
 * \param[in]	propertyName	name of the property to set
 * \param[in]	value	the new value
 * \param[out]	exception	storage for exception data
 * \return	true on success, false on failure
 */
static bool sessionClass_SetProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef value, JSValueRef* exception)
{
	struct objectProperty *session_property = sessionProperties;

	if(!JSObjectGetPrivate(object))
		return true;

	while(session_property && session_property->name) {
		if(JSStringIsEqualToUTF8CString(propertyName, session_property->name) &&
		    session_property->setfunc && !(session_property->flags & kJSPropertyAttributeReadOnly))
		{
			/* found property (and it's writable too) */
			return session_property->setfunc(ctx, object, value, exception);
		}
		session_property++;
	}

	return true;
}

/**
 * sessionClass_DeleteProperty
 * \brief	Delete property handler of session object
 * \param[in]	ctx	context reference
 * \param[in]	object	session object reference
 * \param[in]	propertyName	name of the property to set
 * \param[out]	exception	storage for exception data
 * \return	always true
 */
static bool sessionClass_DeleteProperty(JSContextRef ctx, JSObjectRef object, JSStringRef propertyName, JSValueRef* exception)
{
	/*
	 * We don't delete anything here
	 */
	return true;
}


/**
 * init_session_object
 * \brief	Create session object
 */
static switch_status_t init_session_object(JSContextRef context, switch_core_session_t *session)
{
	struct objectMethod *session_method = sessionMethods;
	JSObjectRef globalObject = JSContextGetGlobalObject(context);
	JSStringRef name, string;
	JSObjectRef object, method;
	JSValueRef value;
	JSSession *js_session;

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Creating session object for session %s\n",
				switch_core_session_get_uuid(session));

	/* session object */
	name = JSStringCreateWithUTF8CString("session");
	object = JSObjectMake(context, globals.sessionClass, NULL);
	JSObjectSetProperty(context, globalObject, name, object, kJSPropertyAttributeNone, NULL);
	JSStringRelease(name);

	/* create JSSession */
	js_session = switch_core_alloc(globals.pool, sizeof(JSSession));
	if(!js_session) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to allocate session handle\n");
		return SWITCH_STATUS_GENERR;
	}
	js_session->session = session;

	/* assign fs core session reference */
	JSObjectSetPrivate(object, js_session);

	/* Setup non-static methods */
	while(session_method && session_method->name) {
		if(session_method->func) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG,
					"Creating non-static method \"%s\" for session object\n",
					session_method->name);

			name = JSStringCreateWithUTF8CString(session_method->name);
			method = JSObjectMakeFunctionWithCallback(context, name, session_method->func);

			/*
			 * important: function properties do not work without kJSPropertyAttributeDontEnum!
			 */
			JSObjectSetProperty(context, object, name, method, kJSPropertyAttributeDontEnum | kJSPropertyAttributeDontDelete | kJSPropertyAttributeReadOnly, NULL);
			JSStringRelease(name);
		}
		session_method++;
	}

	/* Read-Only value properties that never change
	 * during the lifetime of a session (e.g. its UUID)
	 *
	 * Things like "state" will change and are handled
	 * via the Has- / Get- and SetProperty helpers
	 * (and are defined in the "sessionProperties" array)
	 */

	/* uuid */
	name = JSStringCreateWithUTF8CString("uuid");
	string = JSStringCreateWithUTF8CString(switch_core_session_get_uuid(session));
	value  = JSValueMakeString(context, string);
	JSObjectSetProperty(context, object, name, value, kJSPropertyAttributeReadOnly, NULL);
	JSStringRelease(name);
	JSStringRelease(string);

	return SWITCH_STATUS_SUCCESS;
}

/******************************************************************************
 *
 *
 */

/**
 * loadScriptFile
 * \brief	Load scriptfile into newly allocted buffer
 * \param[in]	filename	Path of the script
 * \return	NULL on error, location of buffer otherwise
 */
static char* loadScriptFile(const char* filename)
{
	switch_file_t *fhandle = NULL;
	switch_size_t fsize, nbytes = 1024;
	switch_status_t result;
	char *buffer, *bufptr;

	result = switch_file_open(&fhandle, filename, SWITCH_FOPEN_READ, 0, globals.pool);
	if(result != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Could not open script \"%s\"\n", filename);
		return NULL;
	}

	fsize = switch_file_get_size(fhandle);

	buffer = switch_core_alloc(globals.pool, fsize + 1);
	if(!buffer) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Could not open script \"%s\"\n", filename);
		switch_file_close(fhandle);
		return NULL;
	}

	bufptr = buffer;
	fsize  = 0;

	while((result = switch_file_read(fhandle, bufptr, &nbytes)) == SWITCH_STATUS_SUCCESS) {
		if(nbytes <= 0)
			break;

		bufptr += nbytes;
		fsize  += nbytes;
		nbytes  = 1024;
	}
	buffer[fsize] = '\0';

	switch_file_close(fhandle);

	return buffer;
}

/**
 * log_cb
 * \brief	Logging callback
 */
static JSValueRef log_cb(JSContextRef context, JSObjectRef functionObject, JSObjectRef thisObject,
		 size_t argumentCount,
		 const JSValueRef arguments[],
		 JSValueRef* exception)
{
	if (argumentCount > 0) {
		char tmp[1024];
		JSStringRef message;
		size_t size;
		int loglevel = SWITCH_LOG_NOTICE;


		if(argumentCount == 1) {
			message = JSValueToStringCopy(context, arguments[0], NULL);
		} else {
			JSStringRef level;

			level   = JSValueToStringCopy(context, arguments[0], NULL);
			message = JSValueToStringCopy(context, arguments[1], NULL);

			JSStringGetUTF8CString(level, tmp, sizeof(tmp)-1);

			if(!strcasecmp(tmp, "info")) {
				loglevel = SWITCH_LOG_INFO;
			}
			else if(!strcasecmp(tmp, "notice")) {
				loglevel = SWITCH_LOG_NOTICE;
			}
			else if(!strcasecmp(tmp, "console")) {
				loglevel = SWITCH_LOG_CONSOLE;
			}
			else if(!strcasecmp(tmp, "error")) {
				loglevel = SWITCH_LOG_ERROR;
			}
			else if(!strcasecmp(tmp, "crit")) {
				loglevel = SWITCH_LOG_CRIT;
			}
			else if(!strcasecmp(tmp, "alert")) {
				loglevel = SWITCH_LOG_ALERT;
			}
			else if(!strcasecmp(tmp, "warning")) {
				loglevel = SWITCH_LOG_WARNING;
			}
			else {
				loglevel = SWITCH_LOG_DEBUG;
			}
			JSStringRelease(level);
		}

		size = JSStringGetMaximumUTF8CStringSize(message);
		JSStringGetUTF8CString(message, tmp, ((size < sizeof(tmp)) ? size : sizeof(tmp)));

		switch_log_printf(SWITCH_CHANNEL_LOG, loglevel, "%s\n", tmp);
		JSStringRelease(message);
	}
	return JSValueMakeUndefined(context);
}


/**
 * jsapi_function
 * \brief	FreeSWITCH API function callback
 */
SWITCH_STANDARD_API(jsapi_function)		// outdated
{
	return SWITCH_STATUS_SUCCESS;
}



SWITCH_STANDARD_APP(jsapp_function)
{
	JSGlobalContextRef context;
	JSObjectRef globalObject;
	JSValueRef  exception = NULL, result;
	JSStringRef script;
	switch_status_t retval;
	char tmp[PATH_MAX], *path;

#if 0
	if(!cmd || switch_strlen_zero(cmd)) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Not enough arguments\n");
		return SWITCH_STATUS_GENERR;
	}
#endif
	/* TODO: parse args and extract script filename */
	if(data) {
		path = (char *)data;

		if(*path != '/') {
			snprintf(tmp, sizeof(tmp)-1, "%s%s%s", SWITCH_GLOBAL_dirs.script_dir, SWITCH_PATH_SEPARATOR, path);
			path = tmp;
		}

		if(switch_file_exists(path, NULL) != SWITCH_STATUS_SUCCESS) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Script \"%s\" does not exist\n", path);
			return;
		}
	} else {
		return;
	}

	/*
	 * Create new Squirrelfish context
	 */
#if defined(WEBKIT_VERSION_3_1)
	context = JSGlobalContextCreateInGroup(NULL, NULL);
#else
	context = JSGlobalContextCreate(NULL);
#endif
	globalObject = JSContextGetGlobalObject(context);

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Context and global object created\n");


	/*
	 * Create global functions and classes
	 */
	retval = init_functions(context);
	if(retval != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create global functions\n");
		goto out;
	}

	retval = init_classes(context);
	if(retval != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to create global classes\n");
		goto out;
	}

	/*
	 * Assign pointer to session (or NULL if there is none)
	 */
	if(session) {
		init_session_object(context, session);
	} else {
		script = JSStringCreateWithUTF8CString("var session = null;");
		result = JSEvaluateScript(context, script, NULL, NULL, 1, &exception);

		if(JSValueIsUndefined(context, result))
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "session setting script executed successfully\n");

		JSStringRelease(script);
	}

	/*
	 * Load script
	 */
	char *scriptUTF8 = loadScriptFile(path);
	if(!scriptUTF8) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to load script\n");
		goto out;
	}

	script = JSStringCreateWithUTF8CString(scriptUTF8);
        result = JSEvaluateScript(context, script, NULL, NULL, 1, &exception);
	if(exception) {
		char tmp[1024];
		size_t size;

		JSStringRef exceptionString = JSValueToStringCopy(context, exception, NULL);

		size = JSStringGetMaximumUTF8CStringSize(exceptionString);
		JSStringGetUTF8CString(exceptionString, tmp, ((size < sizeof(tmp)) ? size : sizeof(tmp)));

		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Script failed: %s\n", tmp);

		JSStringRelease(exceptionString);
	} else {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Script executed successfully\n");
	}

	JSStringRelease(script);

out:
	globalObject = NULL;
	JSGarbageCollect(context);
	JSGlobalContextRelease(context);
}

/**
 * mod_squirrelfish_load
 * \brief	FreeSWITCH module interface startup function
 */
SWITCH_MODULE_LOAD_FUNCTION(mod_squirrelfish_load)
{
	switch_api_interface_t *api_interface;
	switch_application_interface_t *app_interface;

	switch_core_new_memory_pool(&globals.pool);

	/* connect my internal structure to the blank pointer passed to me */
	*module_interface = switch_loadable_module_create_module_interface(pool, modname);

	SWITCH_ADD_API(api_interface, "jsrun",      "Run a JavaScript script", jsapi_function, "jsapi <script> [additional_vars [...]]");
	SWITCH_ADD_APP(app_interface, "javascript", "Run a js", "Run a JavaScript script", jsapp_function, "javascript <script> [additional_vars [...]]", SAF_NONE);

	/*
	 * Create Class definitions
	 */
	globals.sessionClass = JSClassCreate(&sessionClass_definition);


	return SWITCH_STATUS_SUCCESS;
}

/**
 * mod_squirrelfish_shutdown
 * \brief	FreeSWITCH module interface shutdown function
 */
SWITCH_MODULE_SHUTDOWN_FUNCTION(mod_squirrelfish_shutdown)
{
	/*
	 * Cleanup routines
	 */

	/*
	 * Free class definitions
	 */
	JSClassRelease(globals.sessionClass);

	return SWITCH_STATUS_SUCCESS;
}

/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:nil
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 expandtab:
 */
